import logging
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import Iterable, Tuple, List, Union, Optional, Iterator

from oaklib.datamodels import obograph
from oaklib.implementations.sparql.sparql_implementation import SparqlImplementation, _sparql_values
from oaklib.implementations.sparql.sparql_query import SparqlQuery
from oaklib.interfaces import SubsetterInterface
from oaklib.interfaces.basic_ontology_interface import RELATIONSHIP_MAP, RELATIONSHIP
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.interfaces.relation_graph_interface import RelationGraphInterface
from oaklib.interfaces.search_interface import SearchInterface
from oaklib.types import CURIE, PRED_CURIE
from oaklib.utilities.graph.networkx_bridge import transitive_reduction_by_predicate
from rdflib import RDFS, RDF, OWL


class RelationGraphEnum(Enum):
    """
    triples in UG are organized into different graphs
    """
    ontology = "http://reasoner.renci.org/ontology"
    redundant = "http://reasoner.renci.org/redundant"
    nonredundant = "http://reasoner.renci.org/nonredundant"


@dataclass
class UbergraphImplementation(SparqlImplementation, RelationGraphInterface, SearchInterface, OboGraphInterface,
                              MappingProviderInterface, SubsetterInterface):
    """
    Wraps the Ubergraph sparql endpoint

    See: `<https://github.com/INCATools/ubergraph>`_

    This is a specialization of the more generic :class:`.SparqlImplementation`, which
    has knowledge of some of the specialized patterns found in Ubergraph

    An UbergraphImplementation can be initialed by:

        .. code:: python

           >>>  oi = UbergraphImplementation.create()

        The default ubergraph endpoint will be assumed

    """
    #sparql_wrapper: SPARQLWrapper

    def _default_url(self) -> str:
        return "https://ubergraph.apps.renci.org/sparql"

    def _is_blazegraph(self) -> bool:
        """
        Currently Ubergraph uses blazegraph
        """
        return True

    @property
    def named_graph(self) -> Optional[str]:
        if not self.resource or self.resource.slug is None:
            return None
        else:
            ont = self.resource.slug
            for g in self.list_of_named_graphs():
                if f'/{ont}.' in g or f'/{ont}-base' in g:
                    return g

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: RelationGraph
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def _get_outgoing_edges_by_curie(self, curie: CURIE, graph: RelationGraphEnum,
                                     predicates: List[PRED_CURIE] = None) -> Iterable[Tuple[CURIE, CURIE]]:
        query_uri = self.curie_to_sparql(curie)
        query = SparqlQuery(select=['?p', '?o'],
                            where=[f'GRAPH <{graph.value}> {{ {query_uri} ?p ?o }}',
                                   f'?o a owl:Class'])
        if predicates:
            pred_uris = [self.curie_to_sparql(pred) for pred in predicates]
            query.where.append(f'VALUES ?p {{ {" ".join(pred_uris)} }}')
        bindings = self._query(query.query_str())
        for row in bindings:
            pred = self.uri_to_curie(row['p']['value'])
            obj = self.uri_to_curie(row['o']['value'])
            yield pred, obj

    def _get_incoming_edges_by_curie(self, curie: CURIE, graph: RelationGraphEnum,
                                     predicates: List[PRED_CURIE] = None) -> Iterable[Tuple[CURIE, CURIE]]:
        query_uri = self.curie_to_sparql(curie)
        query = SparqlQuery(select=['?s', '?p'],
                            where=[f'GRAPH <{graph.value}> {{ ?s ?p {query_uri}  }}',
                                   f'?s a owl:Class'])
        if predicates:
            pred_uris = [self.curie_to_sparql(pred) for pred in predicates]
            query.where.append(f'VALUES ?p {{ {" ".join(pred_uris)} }}')
        bindings = self._query(query.query_str())
        for row in bindings:
            pred = self.uri_to_curie(row['p']['value'])
            subj = self.uri_to_curie(row['s']['value'])
            yield pred, subj


    def get_outgoing_relationships_by_curie(self, curie: CURIE, isa_only: bool = False) -> RELATIONSHIP_MAP:
        rmap = defaultdict(list)
        for pred, obj in self._get_outgoing_edges_by_curie(curie, graph=RelationGraphEnum.nonredundant):
            rmap[pred].append(obj)
        return rmap

    def get_incoming_relationships_by_curie(self, curie: CURIE, isa_only: bool = False) -> RELATIONSHIP_MAP:
        rmap = defaultdict(list)
        for pred, s in self._get_incoming_edges_by_curie(curie, graph=RelationGraphEnum.nonredundant):
            rmap[pred].append(s)
        return rmap

    def entailed_outgoing_relationships_by_curie(self, curie: CURIE,
                                                 predicates: List[PRED_CURIE] = None) -> Iterable[Tuple[PRED_CURIE, CURIE]]:
        return self._get_outgoing_edges_by_curie(curie, graph=RelationGraphEnum.redundant, predicates=predicates)

    def entailed_incoming_relationships_by_curie(self, curie: CURIE,
                                                 predicates: List[PRED_CURIE] = None) -> Iterable[Tuple[PRED_CURIE, CURIE]]:
        return self._get_incoming_edges_by_curie(curie, graph=RelationGraphEnum.redundant, predicates=predicates)

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: OboGraph
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def _values(self, var: str, in_list: Optional[List[str]]) -> str:
        if in_list is None:
            return ""
        else:
            return f'VALUES ?{var} {{ {" ".join(in_list)} }}'

    def _from_subjects_chunked(self, subjects: List[CURIE], predicates: List[PRED_CURIE] = None, **kwargs):
        SIZE = 10
        while len(subjects) > 0:
            next_subjects = subjects[0:SIZE]
            subjects = subjects[SIZE:]
            for r in self._from_subjects(next_subjects, predicates, **kwargs):
                yield r


    def _from_subjects(self, subjects: List[CURIE], predicates: List[PRED_CURIE] = None,
                       graph: str = None, object_is_literal=False, where=[]) -> Iterable[Tuple[CURIE, PRED_CURIE, CURIE]]:
        subject_uris = [self.curie_to_sparql(curie) for curie in subjects]
        if predicates:
            predicate_uris = [self.curie_to_sparql(curie) for curie in predicates]
        else:
            predicate_uris = None
        query = SparqlQuery(select=['?s ?p ?o'],
                            distinct=True,
                            graph=graph,
                            where=['?s ?p ?o',
                                   self._values('s', subject_uris),
                                   self._values('p', predicate_uris),
                                   ] + where)
        #print(f'G={graph} Q={query.query_str()}')
        bindings = self._query(query.query_str())
        for row in bindings:
            v = row['o']['value']
            if not object_is_literal:
                v = self.uri_to_curie(v)
            yield (self.uri_to_curie(row['s']['value']),
                   self.uri_to_curie(row['p']['value']),
                   v)

    def _object_properties(self) -> List[PRED_CURIE]:
        return list(set([t[0] for t in self._triples(None, RDF.type, OWL.ObjectProperty)]))

    def node(self, curie: CURIE) -> obograph.Node:
        params = dict(id=curie,
                      label=self.get_label_by_curie(curie))
        return obograph.Node(**params)

    def ancestor_graph(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> obograph.Graph:
        ancs = list(self.ancestors(start_curies, predicates))
        logging.info(f'NUM ANCS: {len(ancs)}')
        edges = []
        nodes = {}
        for rel in self._from_subjects_chunked(ancs, predicates, graph=RelationGraphEnum.nonredundant.value, where=[]):
            edges.append(obograph.Edge(sub=rel[0], pred=rel[1], obj=rel[2]))
        logging.info(f'NUM EDGES: {len(edges)}')
        for rel in self._from_subjects_chunked(ancs, [RDFS.label], object_is_literal=True):
            id = rel[0]
            nodes[id] = obograph.Node(id=id, label=rel[2])
        logging.info(f'NUM NODES: {len(nodes)}')
        return obograph.Graph(id='query',
                              nodes=list(nodes.values()), edges=edges)

    def relationships_to_graph(self, relationships: Iterable[RELATIONSHIP]) -> obograph.Graph:
        relationships = list(relationships)
        edges = [obograph.Edge(sub=s, pred=p, obj=o) for s, p, o in relationships]
        node_ids = set()
        for rel in relationships:
            node_ids.update(list(rel))
        nodes = {}
        for s, p, o in self._from_subjects_chunked(list(node_ids), [RDFS.label], object_is_literal=True):
            nodes[s] = obograph.Node(id=s, label=o)
        logging.info(f'NUM EDGES: {len(edges)}')
        return obograph.Graph(id='query',
                              nodes=list(nodes.values()), edges=edges)

    def ancestors(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        # TODO: DRY
        query_uris = [self.curie_to_sparql(curie) for curie in start_curies]
        where = [f'?s ?p ?o',
                 f'?o a owl:Class',
                 #f'?p a owl:ObjectProperty',
                 f'VALUES ?s {{ {" ".join(query_uris)} }}']
        if predicates:
            pred_uris = [self.curie_to_sparql(pred) for pred in predicates]
            where.append(f'VALUES ?p {{ {" ".join(pred_uris)} }}')
        query = SparqlQuery(select=['?o'],
                            distinct=True,
                            where=where)
        bindings = self._query(query.query_str())
        for row in bindings:
            yield self.uri_to_curie(row['o']['value'])

    def descendants(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        # TODO: DRY
        query_uris = [self.curie_to_sparql(curie) for curie in start_curies]
        where = [f'?s ?p ?o',
                 f'?s a owl:Class',
                 f'VALUES ?o {{ {" ".join(query_uris)} }}']
        if predicates:
            pred_uris = [self.curie_to_sparql(pred) for pred in predicates]
            where.append(f'VALUES ?p {{ {" ".join(pred_uris)} }}')
        query = SparqlQuery(select=['?s'],
                            distinct=True,
                            where=where)
        bindings = self._query(query.query_str())
        for row in bindings:
            yield self.uri_to_curie(row['s']['value'])

    def gap_fill_relationships(self, seed_curies: List[CURIE], predicates: List[PRED_CURIE] = None) -> Iterator[RELATIONSHIP]:
        # TODO: compare with https://api.triplydb.com/s/_mZ9q_-rg
        query_uris = [self.curie_to_sparql(curie) for curie in seed_curies]
        where = [f'?s ?p ?o',
                 _sparql_values('s', query_uris),
                 _sparql_values('o', query_uris)]
        if predicates:
            pred_uris = [self.curie_to_sparql(pred) for pred in predicates]
            where.append(_sparql_values('p', pred_uris))
        query = SparqlQuery(select=['?s ?p ?o'],
                            where=where)
        bindings = self._query(query.query_str())
        # TODO: remove redundancy
        rels = []
        for row in bindings:
            rels.append( (self.uri_to_curie(row['s']['value']),
                          self.uri_to_curie(row['p']['value']),
                          self.uri_to_curie(row['o']['value'])))
        for rel in transitive_reduction_by_predicate(rels):
            yield rel







# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['oaklib',
 'oaklib.conf',
 'oaklib.datamodels',
 'oaklib.implementations',
 'oaklib.implementations.aggregator',
 'oaklib.implementations.amigo',
 'oaklib.implementations.bioportal',
 'oaklib.implementations.fhir',
 'oaklib.implementations.kgx',
 'oaklib.implementations.neo4j',
 'oaklib.implementations.ols',
 'oaklib.implementations.ontobee',
 'oaklib.implementations.owlery',
 'oaklib.implementations.owlontology',
 'oaklib.implementations.pronto',
 'oaklib.implementations.robot',
 'oaklib.implementations.scigraph',
 'oaklib.implementations.skos',
 'oaklib.implementations.solor',
 'oaklib.implementations.solr',
 'oaklib.implementations.sparql',
 'oaklib.implementations.sqldb',
 'oaklib.implementations.tccm',
 'oaklib.implementations.ubergraph',
 'oaklib.implementations.umls',
 'oaklib.implementations.uniprot',
 'oaklib.implementations.wikidata',
 'oaklib.interfaces',
 'oaklib.io',
 'oaklib.utilities',
 'oaklib.utilities.graph',
 'oaklib.utilities.lexical',
 'oaklib.utilities.mapping',
 'oaklib.utilities.semsim',
 'oaklib.utilities.subsets',
 'oaklib.utilities.taxon']

package_data = \
{'': ['*']}

install_requires = \
['SPARQLWrapper>=2.0.0,<3.0.0',
 'SQLAlchemy>=1.4.32,<2.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'linkml-runtime>=1.2.3,<2.0.0',
 'networkx>=2.7.1,<3.0.0',
 'nxontology>=0.4.0,<0.5.0',
 'pronto>=2.4.4,<3.0.0',
 'ratelimit>=2.2.1,<3.0.0',
 'sssom>=0.3.7,<0.4.0']

entry_points = \
{'console_scripts': ['runoak = oaklib.cli:main']}

setup_kwargs = {
    'name': 'oaklib',
    'version': '0.1.9',
    'description': 'Ontology Access Kit: Python library for common ontology operations over a variety of backends',
    'long_description': '# Ontology Access Kit\n\nPython lib for common ontology operations over a variety of backends.\n\n[![PyPI version](https://badge.fury.io/py/oaklib.svg)](https://badge.fury.io/py/oaklib)\n![](https://github.com/incatools/ontology-access-kit/workflows/Build/badge.svg)\n[![badge](https://img.shields.io/badge/launch-binder-579ACA.svg)](https://mybinder.org/v2/gh/incatools/ontology-access-kit/main?filepath=notebooks)\n\nThis library provides a collection of different [interfaces](https://incatools.github.io/ontology-access-kit/interfaces/index.html) for different kinds of ontology operations, including:\n\n - [lookup of basic features](https://incatools.github.io/ontology-access-kit/interfaces/basic.html) of an ontology element, such as it\'s label, definition, relationships, or aliases\n - search an ontology for a term\n - validating an ontology\n - updating, deleting, or modifying terms\n - ontology term matching\n - generating and visualizing subgraphs\n - provide specialized object models for more advanced operations, such as graph traversal, or OWL axiom processing, or text annotation\n\nThese interfaces are *separated* from any particular [backend](https://incatools.github.io/ontology-access-kit/implementations/index.html). This means the same API can be used regardless of whether the ontology:\n\n - is served by a remote API such as OLS or BioPortal\n - is present locally on the filesystem in owl, obo, obojson, or sqlite formats\n - is to be downloaded from a remote repository such as the OBO library\n - is queried from a remote database, including SPARQL endpoints (Ontobee/Ubergraph), A SQL database, a Solr/ES endpoint\n\n## Documentation:\n\n- [incatools.github.io/ontology-access-kit](https://incatools.github.io/ontology-access-kit)\n\n\n## Example\n\n```python\nresource = OntologyResource(slug=\'tests/input/go-nucleus.db\', local=True)\noi = SqlImplementation(resource)\nfor curie in oi.basic_search("cell"):\n    print(f\'{curie} ! {oi.get_label_by_curie(curie)}\')\n    for rel, fillers in oi.get_outgoing_relationships().items():\n        print(f\'  RELATION: {rel} ! {oi.get_label_by_curie(rel)}\')\n        for filler in fillers:\n            print(f\'     * {filler} ! {oi.get_label_by_curie(filler)}\')\n```\n\nFor more examples, see\n\n- [demo notebook](https://github.com/incatools/ontology-access-kit/blob/main/notebooks/basic-demo.ipynb)\n\n## Command Line\n\nDocumentation here is incomplete.\n\nSee [CLI docs](https://incatools.github.io/ontology-access-kit/cli.html)\n\n### Search\n\nUse the pronto backend to fetch and parse an ontology from the OBO library, then use the `search` command\n\n```bash\nrunoak -i obolibrary:pato.obo search osmol \n```\n\nReturns:\n\n```\nPATO:0001655 ! osmolarity\nPATO:0001656 ! decreased osmolarity\nPATO:0001657 ! increased osmolarity\nPATO:0002027 ! osmolality\nPATO:0002028 ! decreased osmolality\nPATO:0002029 ! increased osmolality\nPATO:0045034 ! normal osmolality\nPATO:0045035 ! normal osmolarity\n```\n\n### QC and Validation\n\nPerform validation on PR using sqlite/rdftab instance:\n\n```bash\nrunoak validate -i sqlite:../semantic-sql/db/pr.db\n```\n\n### List all terms\n\nList all terms obolibrary has for mondo\n\n```bash\nrunoak validate -i obolibrary:mondo.obo terms\n```\n\n### Lexical index\n\nMake a lexical index of all terms in Mondo:\n\n```bash\nrunoak lexmatch -i obolibrary:mondo.obo -L mondo.index.yaml\n```\n\n### Search\n\nSearching over OBO using ontobee:\n\n```bash\nrunoak  -i ontobee: search tentacle\n```\n\nyields:\n\n```\nhttp://purl.obolibrary.org/obo/CEPH_0000256 ! tentacle\nhttp://purl.obolibrary.org/obo/CEPH_0000257 ! tentacle absence\nhttp://purl.obolibrary.org/obo/CEPH_0000258 ! tentacle pad\n...\n```\n\nSearching over a broader set of ontologies in bioportal (requires API KEY)\n\n```bash\nrunoak set-apikey bioportal YOUR-KEY-HERE\nrunoak  -i bioportal: search tentacle\n```\n\nyields:\n\n```\nBTO:0001357 ! tentacle\nhttp://purl.jp/bio/4/id/200906071014668510 ! tentacle\nCEPH:0000256 ! tentacle\nhttp://www.projecthalo.com/aura#Tentacle ! Tentacle\nCEPH:0000256 ! tentacle\n...\n```\n\nSearching over more limited set of ontologies in Ubergraph:\n\n```bash\nrunoak -v -i ubergraph: search tentacle\n```\n\nyields\n```\nUBERON:0013206 ! nasal tentacle\n```\n\n### Annotating Texts\n\n```bash\nrunoak  -i bioportal: annotate neuron from CA4 region of hippocampus of mouse\n```\n\nyields:\n\n```yaml\nobject_id: CL:0000540\nobject_label: neuron\nobject_source: https://data.bioontology.org/ontologies/NIFDYS\nmatch_type: PREF\nsubject_start: 1\nsubject_end: 6\nsubject_label: NEURON\n\nobject_id: http://www.co-ode.org/ontologies/galen#Neuron\nobject_label: Neuron\nobject_source: https://data.bioontology.org/ontologies/GALEN\nmatch_type: PREF\nsubject_start: 1\nsubject_end: 6\nsubject_label: NEURON\n\n...\n```\n\n### Mapping\n\nCreate a SSSOM mapping file for a set of ontologies:\n\n```bash\nrobot merge -I http://purl.obolibrary.org/obo/hp.owl -I http://purl.obolibrary.org/obo/mp.owl convert --check false -o hp-mp.obo\nrunoak lexmatch -i hp-mp.obo -o hp-mp.sssom.tsv\n```\n\n\n\n\n### Visualization of ancestor graphs\n\nUse the sqlite backend to visualize graph up from \'vacuole\' using test ontology sqlite:\n\n```bash\nrunoak -i sqlite:tests/input/go-nucleus.db  viz GO:0005773\n```\n\n![img](notebooks/output/vacuole.png)\n\nSame using ubergraph, restricting to is-a and part-of\n\n```bash\nrunoak -i ubergraph:  viz GO:0005773 -p i,BFO:0000050\n```\n\nSame using pronto, fetching ontology from obolibrary\n\n```bash\nrunoak -i obolibrary:go.obo  viz GO:0005773\n```\n\n## Documentation\n\n- [incatools.github.io/ontology-access-kit](https://incatools.github.io/ontology-access-kit)\n\n## Potential Refactoring\n\nCurrently all implementations exist in this repo/module, this results in a lot of dependencies\n\nOne possibility is to split out each implementation into its own repo and use a plugin architecture\n\n## PyPI release\n\nTODO\n',
    'author': 'cmungall',
    'author_email': 'cjm@berkeleybop.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, Iterator, List, Tuple, Union
from urllib.parse import quote

import requests
import yaml
from oaklib.datamodels.text_annotator import TextAnnotation
from oaklib.interfaces.basic_ontology_interface import PREFIX_MAP
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.search_interface import SearchConfiguration, SearchInterface
from oaklib.interfaces.text_annotator_interface import TextAnnotatorInterface
from oaklib.types import CURIE
from oaklib.utilities.apikey_manager import get_apikey_value
from oaklib.utilities.rate_limiter import check_limit
from sssom import Mapping
from sssom.sssom_datamodel import MatchTypeEnum

REST_URL = "http://data.bioontology.org"

ANNOTATION = Dict[str, Any]

# See: 
#   https://www.bioontology.org/wiki/BioPortal_Mappings 
#   https://github.com/agroportal/project-management/wiki/Mappings
SOURCE_TO_PREDICATE = {
    'CUI': 'skos:closeMatch',
    'LOOM': 'skos:closeMatch',
    'REST': 'skos:relatedMatch', # maybe??
    'SAME_URI': 'skos:exactMatch',
}


@dataclass
class BioportalImplementation(TextAnnotatorInterface, SearchInterface, MappingProviderInterface):
    """
    Implementation over bioportal endpoint

    See `<https://data.bioontology.org/documentation>`_
    """
    bioportal_api_key: str = None
    label_cache: Dict[CURIE, str] = field(default_factory=lambda: {})

    def get_prefix_map(self) -> PREFIX_MAP:
        # TODO
        return {}

    def load_bioportal_api_key(self, path: str = None) -> None:
        self.bioportal_api_key = get_apikey_value('bioportal')

    def _headers(self) -> dict:
        return {'Authorization': 'apikey token=' + self.bioportal_api_key}

    def get_labels_for_curies(self, curies: Iterable[CURIE]) -> Iterable[Tuple[CURIE, str]]:
        label_cache = self.label_cache
        for curie in curies:
            logging.debug(f'LOOKUP: {curie}')
            if curie in label_cache:
                yield curie, label_cache[curie]
            else:
                label = self.get_label_by_curie(curie)
                label_cache[curie] = label
                yield curie, label


    def annotate_text(self, text: str) -> Iterator[TextAnnotation]:
        logging.info(f'Annotating text: {text}')
        #include =['prefLabel', 'synonym', 'definition', 'semanticType', 'cui']
        include =['prefLabel', 'semanticType', 'cui']
        require_exact_match = True
        include_str = ','.join(include)
        params = {'include':  include_str,
                  'require_exact_match': require_exact_match,
                  'text': text}
        if self.bioportal_api_key is  None:
            self.load_bioportal_api_key()
        check_limit()
        r = requests.get(REST_URL + '/annotator',
                         headers=self._headers(),
                         params=params)
        return self.json_to_results(r.json(), text)

    def json_to_results(self, json_list: List[Any], text: str) -> Iterator[TextAnnotation]:
        results = []
        seen = {}
        for obj in json_list:
            ac_obj = obj['annotatedClass']
            for x in obj['annotations']:
                ann = TextAnnotation(subject_start=x['from'],
                                     subject_end=x['to'],
                                     subject_label=x['text'],
                                     object_id=self.uri_to_curie(ac_obj['@id']),
                                     object_label=ac_obj['prefLabel'],
                                     object_source=ac_obj['links']['ontology'],
                                     match_type=x['matchType'],
                                     #info=str(obj)
                                     )
                uid = ann.subject_start, ann.subject_end, ann.object_id
                if uid in seen:
                    logging.debug(f'Skipping duplicative annotation to {ann.object_source}')
                    continue
                seen[uid] = True
                yield ann

    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: SearchInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


    def basic_search(self, search_term: str, config: SearchConfiguration = SearchConfiguration()) -> Iterable[CURIE]:
        if self.bioportal_api_key is  None:
            self.load_bioportal_api_key()
        check_limit()
        r = requests.get(REST_URL + '/search',
                         headers=self._headers(),
                         params={'q': search_term, 'include': ['prefLabel']})
        obj = r.json()
        collection = obj['collection']
        while len(collection) > 0:
            result = collection[0]
            curie = self.uri_to_curie(result['@id'])
            label = result.get('prefLabel', None)
            self.label_cache[curie] = label
            logging.debug(f'M: {curie} => {label}')
            yield curie
            collection = collection[1:]
            if len(collection) == 0:
                next_page = obj['links']['nextPage']
                #print(f'NEXT={next_page}')
                if next_page:
                    check_limit()
                    r = requests.get(next_page, headers=self._headers())
                    obj = r.json()
                    collection = obj['collection']


    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    # Implements: MappingProviderInterface
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    def get_sssom_mappings_by_curie(self, curie: CURIE) -> Iterable[Mapping]:
        if self.bioportal_api_key is  None:
            self.load_bioportal_api_key()
        [prefix, _] = curie.split(':', 2)
        class_uri = quote(self.curie_to_uri(curie), safe='')
        # This may return lots of duplicate mappings
        # See: https://github.com/ncbo/ontologies_linked_data/issues/117
        req_url = f'{REST_URL}/ontologies/{prefix}/classes/{class_uri}/mappings'
        response = requests.get(req_url, headers=self._headers(), params={'display_links': 'false', 'display_context': 'false'})
        body = response.json()
        for result in body:
            yield self.result_to_mapping(result)


    def result_to_mapping(self, result: Dict[str, Any]) -> Mapping:
        mapping = Mapping(
            subject_id=result['classes'][0]['@id'],
            predicate_id=SOURCE_TO_PREDICATE[result['source']],
            match_type=MatchTypeEnum.Unspecified,
            object_id=result['classes'][1]['@id'],
            mapping_provider=result['@type'],
            mapping_tool=result['source'],
        )
        return mapping



from collections import defaultdict
from typing import Tuple, Any, Dict, List, Iterable


def pairs_as_dict(pairs: Iterable[Tuple[Any, Any]]) -> Dict[Any, List[Any]]:
    d = defaultdict(list)
    for p in pairs:
        d[p[0]].append(d[1])
    return d


from abc import ABC
from dataclasses import dataclass
from typing import Dict, List, Tuple, Any

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface, RELATIONSHIP_MAP, RELATIONSHIP
#import oaklib.datamodels.summary_statistics_datamodel.slots as statdict_slots
from oaklib.types import CURIE, LABEL, URI, PRED_CURIE


SUMMARY_STATISTICS_MAP = Dict[str, Any]


class SummaryStatisticsInterface(BasicOntologyInterface, ABC):
    """
    generates summary statistics

    Uses SummaryStatistics datamodel, see :ref:`datamodels`
    """

    def get_global_summary_statistics(self) -> SUMMARY_STATISTICS_MAP:
        """
        Gets summary statistics for all ontologies wrapped as a single dictionary

        :return:
        """
        raise NotImplementedError

    def get_partitioned_summary_statistics(self) -> Dict[CURIE, SUMMARY_STATISTICS_MAP]:
        """
        Gets summary statistics for all ontologies wrapped as multiple dictionaries keyed
        by ontology CURIE

        :return:
        """
        raise NotImplementedError

    def get_metric_description(self, metric: str) -> str:
        """
        Examines the data dictionary to retrieve the definition of a metric

        Any dictionary can be used, but summary_statistics_datamodel is favored

        :param metric:
        :return:
        """
        raise NotImplementedError




# Auto generated from obograph.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-03-31T11:02:39
# Schema: obographs_datamodel
#
# id: https://github.com/geneontology/obographs
# description: Schema for benchmarking based on obographs
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, String
from linkml_runtime.utils.metamodelcore import Bool

metamodel_version = "1.7.0"
version = "0.0.1"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
OG = CurieNamespace('og', 'https://github.com/geneontology/obographs/')
RDF = CurieNamespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = CurieNamespace('rdfs', 'http://www.w3.org/2000/01/rdf-schema#')
SDO = CurieNamespace('sdo', 'https://schema.org/')
SKOS = CurieNamespace('skos', 'http://www.w3.org/2004/02/skos/core#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = OG


# Types

# Class references
class GraphId(extended_str):
    pass


class NodeId(extended_str):
    pass


@dataclass
class Graph(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.Graph
    class_class_curie: ClassVar[str] = "og:Graph"
    class_name: ClassVar[str] = "Graph"
    class_model_uri: ClassVar[URIRef] = OG.Graph

    id: Union[str, GraphId] = None
    lbl: Optional[str] = None
    meta: Optional[Union[dict, "Meta"]] = None
    nodes: Optional[Union[Dict[Union[str, NodeId], Union[dict, "Node"]], List[Union[dict, "Node"]]]] = empty_dict()
    edges: Optional[Union[Union[dict, "Edge"], List[Union[dict, "Edge"]]]] = empty_list()
    equivalentNodesSets: Optional[Union[Union[dict, "EquivalentNodesSet"], List[Union[dict, "EquivalentNodesSet"]]]] = empty_list()
    logicalDefinitionAxioms: Optional[Union[Union[dict, "LogicalDefinitionAxiom"], List[Union[dict, "LogicalDefinitionAxiom"]]]] = empty_list()
    domainRangeAxioms: Optional[Union[Union[dict, "DomainRangeAxiom"], List[Union[dict, "DomainRangeAxiom"]]]] = empty_list()
    propertyChainAxioms: Optional[Union[Union[dict, "PropertyChainAxiom"], List[Union[dict, "PropertyChainAxiom"]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, GraphId):
            self.id = GraphId(self.id)

        if self.lbl is not None and not isinstance(self.lbl, str):
            self.lbl = str(self.lbl)

        if self.meta is not None and not isinstance(self.meta, Meta):
            self.meta = Meta(**as_dict(self.meta))

        self._normalize_inlined_as_list(slot_name="nodes", slot_type=Node, key_name="id", keyed=True)

        if not isinstance(self.edges, list):
            self.edges = [self.edges] if self.edges is not None else []
        self.edges = [v if isinstance(v, Edge) else Edge(**as_dict(v)) for v in self.edges]

        if not isinstance(self.equivalentNodesSets, list):
            self.equivalentNodesSets = [self.equivalentNodesSets] if self.equivalentNodesSets is not None else []
        self.equivalentNodesSets = [v if isinstance(v, EquivalentNodesSet) else EquivalentNodesSet(**as_dict(v)) for v in self.equivalentNodesSets]

        if not isinstance(self.logicalDefinitionAxioms, list):
            self.logicalDefinitionAxioms = [self.logicalDefinitionAxioms] if self.logicalDefinitionAxioms is not None else []
        self.logicalDefinitionAxioms = [v if isinstance(v, LogicalDefinitionAxiom) else LogicalDefinitionAxiom(**as_dict(v)) for v in self.logicalDefinitionAxioms]

        if not isinstance(self.domainRangeAxioms, list):
            self.domainRangeAxioms = [self.domainRangeAxioms] if self.domainRangeAxioms is not None else []
        self.domainRangeAxioms = [v if isinstance(v, DomainRangeAxiom) else DomainRangeAxiom(**as_dict(v)) for v in self.domainRangeAxioms]

        if not isinstance(self.propertyChainAxioms, list):
            self.propertyChainAxioms = [self.propertyChainAxioms] if self.propertyChainAxioms is not None else []
        self.propertyChainAxioms = [v if isinstance(v, PropertyChainAxiom) else PropertyChainAxiom(**as_dict(v)) for v in self.propertyChainAxioms]

        super().__post_init__(**kwargs)


@dataclass
class Node(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.Node
    class_class_curie: ClassVar[str] = "og:Node"
    class_name: ClassVar[str] = "Node"
    class_model_uri: ClassVar[URIRef] = OG.Node

    id: Union[str, NodeId] = None
    label: Optional[str] = None
    type: Optional[str] = None
    meta: Optional[Union[dict, "Meta"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, NodeId):
            self.id = NodeId(self.id)

        if self.label is not None and not isinstance(self.label, str):
            self.label = str(self.label)

        if self.type is not None and not isinstance(self.type, str):
            self.type = str(self.type)

        if self.meta is not None and not isinstance(self.meta, Meta):
            self.meta = Meta(**as_dict(self.meta))

        super().__post_init__(**kwargs)


@dataclass
class Edge(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.Edge
    class_class_curie: ClassVar[str] = "og:Edge"
    class_name: ClassVar[str] = "Edge"
    class_model_uri: ClassVar[URIRef] = OG.Edge

    sub: Optional[str] = None
    pred: Optional[str] = None
    obj: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.sub is not None and not isinstance(self.sub, str):
            self.sub = str(self.sub)

        if self.pred is not None and not isinstance(self.pred, str):
            self.pred = str(self.pred)

        if self.obj is not None and not isinstance(self.obj, str):
            self.obj = str(self.obj)

        super().__post_init__(**kwargs)


@dataclass
class Meta(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.Meta
    class_class_curie: ClassVar[str] = "og:Meta"
    class_name: ClassVar[str] = "Meta"
    class_model_uri: ClassVar[URIRef] = OG.Meta

    subsets: Optional[Union[str, List[str]]] = empty_list()
    version: Optional[str] = None
    comments: Optional[Union[str, List[str]]] = empty_list()
    definition: Optional[Union[dict, "DefinitionPropertyValue"]] = None
    xrefs: Optional[Union[str, List[str]]] = empty_list()
    synonyms: Optional[Union[Union[dict, "SynonymPropertyValue"], List[Union[dict, "SynonymPropertyValue"]]]] = empty_list()
    basicPropertyValues: Optional[Union[Union[dict, "BasicPropertyValue"], List[Union[dict, "BasicPropertyValue"]]]] = empty_list()
    deprecated: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.subsets, list):
            self.subsets = [self.subsets] if self.subsets is not None else []
        self.subsets = [v if isinstance(v, str) else str(v) for v in self.subsets]

        if self.version is not None and not isinstance(self.version, str):
            self.version = str(self.version)

        if not isinstance(self.comments, list):
            self.comments = [self.comments] if self.comments is not None else []
        self.comments = [v if isinstance(v, str) else str(v) for v in self.comments]

        if self.definition is not None and not isinstance(self.definition, DefinitionPropertyValue):
            self.definition = DefinitionPropertyValue(**as_dict(self.definition))

        if not isinstance(self.xrefs, list):
            self.xrefs = [self.xrefs] if self.xrefs is not None else []
        self.xrefs = [v if isinstance(v, str) else str(v) for v in self.xrefs]

        if not isinstance(self.synonyms, list):
            self.synonyms = [self.synonyms] if self.synonyms is not None else []
        self.synonyms = [v if isinstance(v, SynonymPropertyValue) else SynonymPropertyValue(**as_dict(v)) for v in self.synonyms]

        if not isinstance(self.basicPropertyValues, list):
            self.basicPropertyValues = [self.basicPropertyValues] if self.basicPropertyValues is not None else []
        self.basicPropertyValues = [v if isinstance(v, BasicPropertyValue) else BasicPropertyValue(**as_dict(v)) for v in self.basicPropertyValues]

        if self.deprecated is not None and not isinstance(self.deprecated, Bool):
            self.deprecated = Bool(self.deprecated)

        super().__post_init__(**kwargs)


@dataclass
class PropertyValue(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.PropertyValue
    class_class_curie: ClassVar[str] = "og:PropertyValue"
    class_name: ClassVar[str] = "PropertyValue"
    class_model_uri: ClassVar[URIRef] = OG.PropertyValue

    pred: Optional[str] = None
    val: Optional[str] = None
    xrefs: Optional[Union[str, List[str]]] = empty_list()
    meta: Optional[Union[dict, Meta]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.pred is not None and not isinstance(self.pred, str):
            self.pred = str(self.pred)

        if self.val is not None and not isinstance(self.val, str):
            self.val = str(self.val)

        if not isinstance(self.xrefs, list):
            self.xrefs = [self.xrefs] if self.xrefs is not None else []
        self.xrefs = [v if isinstance(v, str) else str(v) for v in self.xrefs]

        if self.meta is not None and not isinstance(self.meta, Meta):
            self.meta = Meta(**as_dict(self.meta))

        super().__post_init__(**kwargs)


class DefinitionPropertyValue(PropertyValue):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.DefinitionPropertyValue
    class_class_curie: ClassVar[str] = "og:DefinitionPropertyValue"
    class_name: ClassVar[str] = "DefinitionPropertyValue"
    class_model_uri: ClassVar[URIRef] = OG.DefinitionPropertyValue


class BasicPropertyValue(PropertyValue):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.BasicPropertyValue
    class_class_curie: ClassVar[str] = "og:BasicPropertyValue"
    class_name: ClassVar[str] = "BasicPropertyValue"
    class_model_uri: ClassVar[URIRef] = OG.BasicPropertyValue


class XrefPropertyValue(PropertyValue):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.XrefPropertyValue
    class_class_curie: ClassVar[str] = "og:XrefPropertyValue"
    class_name: ClassVar[str] = "XrefPropertyValue"
    class_model_uri: ClassVar[URIRef] = OG.XrefPropertyValue


@dataclass
class SynonymPropertyValue(PropertyValue):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.SynonymPropertyValue
    class_class_curie: ClassVar[str] = "og:SynonymPropertyValue"
    class_name: ClassVar[str] = "SynonymPropertyValue"
    class_model_uri: ClassVar[URIRef] = OG.SynonymPropertyValue

    synonymType: Optional[str] = None
    isExact: Optional[Union[bool, Bool]] = None
    scope: Optional[Union[str, "ScopesEnum"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.synonymType is not None and not isinstance(self.synonymType, str):
            self.synonymType = str(self.synonymType)

        if self.isExact is not None and not isinstance(self.isExact, Bool):
            self.isExact = Bool(self.isExact)

        if self.scope is not None and not isinstance(self.scope, ScopesEnum):
            self.scope = ScopesEnum(self.scope)

        super().__post_init__(**kwargs)


@dataclass
class Axiom(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.Axiom
    class_class_curie: ClassVar[str] = "og:Axiom"
    class_name: ClassVar[str] = "Axiom"
    class_model_uri: ClassVar[URIRef] = OG.Axiom

    meta: Optional[Union[dict, Meta]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.meta is not None and not isinstance(self.meta, Meta):
            self.meta = Meta(**as_dict(self.meta))

        super().__post_init__(**kwargs)


class DomainRangeAxiom(Axiom):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.DomainRangeAxiom
    class_class_curie: ClassVar[str] = "og:DomainRangeAxiom"
    class_name: ClassVar[str] = "DomainRangeAxiom"
    class_model_uri: ClassVar[URIRef] = OG.DomainRangeAxiom


@dataclass
class EquivalentNodesSet(Axiom):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.EquivalentNodesSet
    class_class_curie: ClassVar[str] = "og:EquivalentNodesSet"
    class_name: ClassVar[str] = "EquivalentNodesSet"
    class_model_uri: ClassVar[URIRef] = OG.EquivalentNodesSet

    representitiveNodeId: Optional[str] = None
    nodeIds: Optional[Union[str, List[str]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.representitiveNodeId is not None and not isinstance(self.representitiveNodeId, str):
            self.representitiveNodeId = str(self.representitiveNodeId)

        if not isinstance(self.nodeIds, list):
            self.nodeIds = [self.nodeIds] if self.nodeIds is not None else []
        self.nodeIds = [v if isinstance(v, str) else str(v) for v in self.nodeIds]

        super().__post_init__(**kwargs)


@dataclass
class ExistentialRestrictionExpression(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.ExistentialRestrictionExpression
    class_class_curie: ClassVar[str] = "og:ExistentialRestrictionExpression"
    class_name: ClassVar[str] = "ExistentialRestrictionExpression"
    class_model_uri: ClassVar[URIRef] = OG.ExistentialRestrictionExpression

    fillerId: Optional[str] = None
    propertyId: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.fillerId is not None and not isinstance(self.fillerId, str):
            self.fillerId = str(self.fillerId)

        if self.propertyId is not None and not isinstance(self.propertyId, str):
            self.propertyId = str(self.propertyId)

        super().__post_init__(**kwargs)


class LogicalDefinitionAxiom(Axiom):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.LogicalDefinitionAxiom
    class_class_curie: ClassVar[str] = "og:LogicalDefinitionAxiom"
    class_name: ClassVar[str] = "LogicalDefinitionAxiom"
    class_model_uri: ClassVar[URIRef] = OG.LogicalDefinitionAxiom


class PropertyChainAxiom(Axiom):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OG.PropertyChainAxiom
    class_class_curie: ClassVar[str] = "og:PropertyChainAxiom"
    class_name: ClassVar[str] = "PropertyChainAxiom"
    class_model_uri: ClassVar[URIRef] = OG.PropertyChainAxiom


# Enumerations
class ScopesEnum(EnumDefinitionImpl):

    exact = PermissibleValue(text="exact")
    narrow = PermissibleValue(text="narrow")
    broad = PermissibleValue(text="broad")
    related = PermissibleValue(text="related")

    _defn = EnumDefinition(
        name="ScopesEnum",
    )

class PredsEnum(EnumDefinitionImpl):

    hasExactSynonym = PermissibleValue(text="hasExactSynonym")
    hasNarrowSynonym = PermissibleValue(text="hasNarrowSynonym")
    hasBroadSynonym = PermissibleValue(text="hasBroadSynonym")
    hasRelatedSynonym = PermissibleValue(text="hasRelatedSynonym")

    _defn = EnumDefinition(
        name="PredsEnum",
    )

# Slots
class slots:
    pass

slots.id = Slot(uri=OG.id, name="id", curie=OG.curie('id'),
                   model_uri=OG.id, domain=None, range=URIRef)

slots.sub = Slot(uri=OG.sub, name="sub", curie=OG.curie('sub'),
                   model_uri=OG.sub, domain=None, range=Optional[str])

slots.pred = Slot(uri=OG.pred, name="pred", curie=OG.curie('pred'),
                   model_uri=OG.pred, domain=None, range=Optional[str])

slots.obj = Slot(uri=OG.obj, name="obj", curie=OG.curie('obj'),
                   model_uri=OG.obj, domain=None, range=Optional[str])

slots.val = Slot(uri=OG.val, name="val", curie=OG.curie('val'),
                   model_uri=OG.val, domain=None, range=Optional[str])

slots.label = Slot(uri=OG.label, name="label", curie=OG.curie('label'),
                   model_uri=OG.label, domain=None, range=Optional[str])

slots.type = Slot(uri=OG.type, name="type", curie=OG.curie('type'),
                   model_uri=OG.type, domain=None, range=Optional[str])

slots.meta = Slot(uri=OG.meta, name="meta", curie=OG.curie('meta'),
                   model_uri=OG.meta, domain=None, range=Optional[Union[dict, Meta]])

slots.definition = Slot(uri=OG.definition, name="definition", curie=OG.curie('definition'),
                   model_uri=OG.definition, domain=None, range=Optional[Union[dict, DefinitionPropertyValue]])

slots.basicPropertyValues = Slot(uri=OG.basicPropertyValues, name="basicPropertyValues", curie=OG.curie('basicPropertyValues'),
                   model_uri=OG.basicPropertyValues, domain=None, range=Optional[Union[Union[dict, BasicPropertyValue], List[Union[dict, BasicPropertyValue]]]])

slots.comments = Slot(uri=OG.comments, name="comments", curie=OG.curie('comments'),
                   model_uri=OG.comments, domain=None, range=Optional[Union[str, List[str]]])

slots.version = Slot(uri=OG.version, name="version", curie=OG.curie('version'),
                   model_uri=OG.version, domain=None, range=Optional[str])

slots.deprecated = Slot(uri=OG.deprecated, name="deprecated", curie=OG.curie('deprecated'),
                   model_uri=OG.deprecated, domain=None, range=Optional[Union[bool, Bool]])

slots.subsets = Slot(uri=OG.subsets, name="subsets", curie=OG.curie('subsets'),
                   model_uri=OG.subsets, domain=None, range=Optional[Union[str, List[str]]])

slots.xrefs = Slot(uri=OG.xrefs, name="xrefs", curie=OG.curie('xrefs'),
                   model_uri=OG.xrefs, domain=None, range=Optional[Union[str, List[str]]])

slots.lbl = Slot(uri=OG.lbl, name="lbl", curie=OG.curie('lbl'),
                   model_uri=OG.lbl, domain=None, range=Optional[str])

slots.nodes = Slot(uri=OG.nodes, name="nodes", curie=OG.curie('nodes'),
                   model_uri=OG.nodes, domain=None, range=Optional[Union[Dict[Union[str, NodeId], Union[dict, Node]], List[Union[dict, Node]]]])

slots.edges = Slot(uri=OG.edges, name="edges", curie=OG.curie('edges'),
                   model_uri=OG.edges, domain=None, range=Optional[Union[Union[dict, Edge], List[Union[dict, Edge]]]])

slots.equivalentNodesSets = Slot(uri=OG.equivalentNodesSets, name="equivalentNodesSets", curie=OG.curie('equivalentNodesSets'),
                   model_uri=OG.equivalentNodesSets, domain=None, range=Optional[Union[Union[dict, EquivalentNodesSet], List[Union[dict, EquivalentNodesSet]]]])

slots.logicalDefinitionAxioms = Slot(uri=OG.logicalDefinitionAxioms, name="logicalDefinitionAxioms", curie=OG.curie('logicalDefinitionAxioms'),
                   model_uri=OG.logicalDefinitionAxioms, domain=None, range=Optional[Union[Union[dict, LogicalDefinitionAxiom], List[Union[dict, LogicalDefinitionAxiom]]]])

slots.domainRangeAxioms = Slot(uri=OG.domainRangeAxioms, name="domainRangeAxioms", curie=OG.curie('domainRangeAxioms'),
                   model_uri=OG.domainRangeAxioms, domain=None, range=Optional[Union[Union[dict, DomainRangeAxiom], List[Union[dict, DomainRangeAxiom]]]])

slots.propertyChainAxioms = Slot(uri=OG.propertyChainAxioms, name="propertyChainAxioms", curie=OG.curie('propertyChainAxioms'),
                   model_uri=OG.propertyChainAxioms, domain=None, range=Optional[Union[Union[dict, PropertyChainAxiom], List[Union[dict, PropertyChainAxiom]]]])

slots.representitiveNodeId = Slot(uri=OG.representitiveNodeId, name="representitiveNodeId", curie=OG.curie('representitiveNodeId'),
                   model_uri=OG.representitiveNodeId, domain=None, range=Optional[str])

slots.nodeIds = Slot(uri=OG.nodeIds, name="nodeIds", curie=OG.curie('nodeIds'),
                   model_uri=OG.nodeIds, domain=None, range=Optional[Union[str, List[str]]])

slots.fillerId = Slot(uri=OG.fillerId, name="fillerId", curie=OG.curie('fillerId'),
                   model_uri=OG.fillerId, domain=None, range=Optional[str])

slots.propertyId = Slot(uri=OG.propertyId, name="propertyId", curie=OG.curie('propertyId'),
                   model_uri=OG.propertyId, domain=None, range=Optional[str])

slots.synonyms = Slot(uri=OG.synonyms, name="synonyms", curie=OG.curie('synonyms'),
                   model_uri=OG.synonyms, domain=None, range=Optional[Union[Union[dict, SynonymPropertyValue], List[Union[dict, SynonymPropertyValue]]]])

slots.synonymType = Slot(uri=OG.synonymType, name="synonymType", curie=OG.curie('synonymType'),
                   model_uri=OG.synonymType, domain=None, range=Optional[str])

slots.scope = Slot(uri=OG.scope, name="scope", curie=OG.curie('scope'),
                   model_uri=OG.scope, domain=None, range=Optional[Union[str, "ScopesEnum"]])

slots.isExact = Slot(uri=OG.isExact, name="isExact", curie=OG.curie('isExact'),
                   model_uri=OG.isExact, domain=None, range=Optional[Union[bool, Bool]])



"""
Utilities for working with ontology subsets (slims)
----
"""

import logging
from typing import List, Dict

from oaklib.interfaces import RelationGraphInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.types import CURIE, PRED_CURIE

def filter_redundant(oi: RelationGraphInterface, curies: List[CURIE], predicates: List[PRED_CURIE] = None) -> List[CURIE]:
    return [curie for curie in curies if not is_redundant(oi, curie, curies, predicates)]

def is_redundant(oi: RelationGraphInterface, curie: CURIE, curies: List[CURIE], predicates: List[PRED_CURIE] = None) -> bool:
    for candidate in curies:
        if candidate != curie:
            if curie in list(oi.ancestors(candidate, predicates=predicates)):
                return True
    return False

def roll_up_to_named_subset(oi: OboGraphInterface, subset: CURIE, curies: List[CURIE],
                            predicates: List[PRED_CURIE] = None) -> Dict[CURIE, List[CURIE]]:
    """
    Rolls up all specified curies to a named subset, e.g. goslim_generic

    :param oi:
    :param subset:
    :param curies:
    :param predicates:
    :return:
    """
    # see https://metacpan.org/dist/go-perl/view/scripts/map2slim
    terms_in_subset = set(oi.curies_by_subset(subset))
    #print(f'SUBSET={list(terms_in_subset)}')
    logging.info(f'Terms in {subset} = {len(terms_in_subset)}')
    subset_anc_map = {t: [a for a in oi.ancestors(t, predicates)
                          if a != t and a in terms_in_subset] for t in terms_in_subset}
    m = {}
    for curie in curies:
        logging.info(f'  Mapping {curie}')
        ancs = list(oi.ancestors(curie, predicates=predicates))
        #print(f'{curie} ANCS={list(ancs)}')
        ancs_in_subset = terms_in_subset.intersection(ancs)
        #print(f'    xx ANCS={list(ancs_in_subset)}')
        redundant = set()
        for a in ancs_in_subset:
            redundant.update(subset_anc_map[a])
        nr_ancs_in_subset = [a for a in ancs_in_subset if a not in redundant]
        m[curie] = nr_ancs_in_subset
    return m







import logging
from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Tuple, Iterable, Union, Iterator, Optional, Any

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface, RELATIONSHIP_MAP, RELATIONSHIP
from oaklib.types import CURIE, LABEL, URI, PRED_CURIE
from oaklib.utilities.graph.relationship_walker import walk_up, walk_down
from oaklib.datamodels.obograph import Node, Graph, Edge

class Distance(Enum):
    """
    Specifies how many hops to walk in any given direction
    """
    ZERO = "zero"
    DIRECT = "direct"
    TRANSITIVE = "transitive"


@dataclass
class TraversalConfiguration:
    """
    Specifies how to walk up and down a graph
    """
    predicates: List[PRED_CURIE] = None
    up_distance: Distance = field(default_factory=lambda: Distance.TRANSITIVE)
    down_distance: Distance = field(default_factory=lambda: Distance.TRANSITIVE)


class OboGraphInterface(BasicOntologyInterface, ABC):
    """
    an interface that provides an Object Oriented view of an ontology, following the OBO Graph Datamodel

    See `OBOGraphs <https://github.com/geneontology/obographs>`_

    Key datamodel concepts:

    - :class:`obograph.Node` - any named ontology element
    - :class:`obograph.Edge` - any relationship between elements; for example between "finger" and "hand"
    - :class:`obograph.Graph` - a collection of nodes, edges, and other ontology components

    This datamodel conceives of an ontology as a graph
    """
    transitive_query_cache: Dict[Any, Any] = None

    def enable_transitive_query_cache(self):
        """
        Cache transitive queries
        """
        self.transitive_query_cache = {}

    def disable_transitive_query_cache(self):
        """
        Do not cache transitive queries (default)
        """
        self.transitive_query_cache = None

    def nodes(self) -> Iterator[Node]:
        """
        Iterator over all nodes in all graphs

        :return:
        """
        raise NotImplementedError

    def edges(self) -> Iterator[Edge]:
        """
        Iterator over all edges in all graphs

        :return:
        """
        raise NotImplementedError

    def node(self, curie: CURIE, strict=False) -> Node:
        """
        Look up a node object by CURIE

        :param curie:
        :param strict:
        :return:
        """
        raise NotImplementedError

    def _graph(self, triples: Iterable[RELATIONSHIP]) -> Graph:
        node_map: Dict[str, Node] = {}
        edges = []
        for s, p, o in triples:
            if s not in node_map:
                node_map[s] = self.node(s)
            if p not in node_map:
                node_map[p] = self.node(p)
            if o not in node_map:
                node_map[o] = self.node(o)
            edges.append(Edge(sub=s, pred=p, obj=o))
        graph_id = 'test'
        return Graph(id=graph_id,
                     nodes=list(node_map.values()),
                     edges=edges)

    def ancestor_graph(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Graph:
        """
        Return a graph object that consists of all the nodes specified in the start_curies list,
        extended with an interactive walk up the graph following all relationships (optionally filtered by the predicate
        list)

        :param start_curies:
        :param predicates: if supplied then only follow edges with these predicates
        :return: ancestor graph
        """
        key = ('ancestor_graph', tuple(start_curies), tuple(predicates if predicates is not None else ()))
        if self.transitive_query_cache is not None:
            if key in self.transitive_query_cache:
                return self.transitive_query_cache[key]
        g = self._graph(walk_up(self, start_curies, predicates=predicates))
        if self.transitive_query_cache is not None:
            self.transitive_query_cache[key] = g
        return g

    def descendant_graph(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Graph:
        """
        As ancestor graph, but in opposite direction

        :param start_curies:
        :param predicates: if supplied then only follow edges with these predicates
        :return: ancestor graph
        """
        key = ('descendant_graph', tuple(start_curies), tuple(predicates if predicates is not None else ()))
        if self.transitive_query_cache is not None:
            if key in self.transitive_query_cache:
                return self.transitive_query_cache[key]
        g = self._graph(walk_down(self, start_curies, predicates=predicates))
        if self.transitive_query_cache is not None:
            self.transitive_query_cache[key] = g
        return g

    def ancestors(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        """
        Ancestors obtained from a walk starting from start_curies ending in roots, following only the specified
        predicates.

        .. note::

           This operation is reflexive: self is included

        :param start_curies: curie or curies to start the walk from
        :param predicates: only traverse over these (traverses over all if this is not set)
        :return: all ancestor CURIEs
        """
        for node in self.ancestor_graph(start_curies, predicates).nodes:
            yield node.id

    def descendants(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None) -> Iterable[CURIE]:
        """
        Descendants obtained from a walk downwards starting from start_curies ending in roots, following only the specified
        predicates.

        .. note::

           This operation is reflexive: self is included

        :param start_curies: curie or curies to start the walk from
        :param predicates: only traverse over these (traverses over all if this is not set)
        :return: all descendant CURIEs
        """
        for node in self.descendant_graph(start_curies, predicates).nodes:
            yield node.id

    def subgraph(self, start_curies: Union[CURIE, List[CURIE]], predicates: List[PRED_CURIE] = None,
                 traversal: TraversalConfiguration = None) -> Graph:
        """
        Combines ancestors and descendants according to a traversal configuration

        :param start_curies:
        :param predicates:
        :param traversal:
        :return:
        """
        if traversal is None:
            traversal = TraversalConfiguration()
        if traversal.up_distance == Distance.TRANSITIVE:
            logging.info(f'Getting ancestor graph from {type(self)}, start={start_curies}')
            up_graph = self.ancestor_graph(start_curies, predicates=predicates)
        else:
            up_graph = None
        if traversal.down_distance == Distance.TRANSITIVE:
            down_graph = self.descendant_graph(start_curies, predicates=predicates)
        else:
            down_graph = None
        g = self._merge_graphs([up_graph, down_graph])
        return g

    def relationships_to_graph(self, relationships: Iterable[RELATIONSHIP]) -> Graph:
        """
        Generates an OboGraph from a list of relationships

        :param relationships:
        :return:
        """
        relationships = list(relationships)
        node_ids = set()
        for rel in relationships:
            node_ids.update(list(rel))
        edges = [Edge(sub=s, pred=p, obj=o) for s, p, o in relationships]
        nodes = [self.node(id) for id in node_ids]
        return Graph(id='query',
                     nodes=list(nodes), edges=edges)

    def as_obograph(self) -> Graph:
        """
        Convert entire resource to an OBO Graph object

        .. warning ::

           some remote resources may choose to throw a NotImplementedError if it is impractical
           to download the entire ontology as a graph
        :return:
        """
        raise NotImplementedError

    def _merge_graphs(self, graphs: List[Optional[Graph]]) -> Graph:
        g = Graph(id='merged')
        for src in graphs:
            if src is not None:
                g.nodes += src.nodes
        for src in graphs:
            if src is not None:
                g.edges += src.edges
        return g












from abc import ABC
from dataclasses import dataclass
from typing import Dict, List, Iterable, Tuple, Optional, Any, Iterator

from oaklib.interfaces.ontology_interface import OntologyInterface
from oaklib.types import CURIE, URI, PRED_CURIE, SUBSET_CURIE
from oaklib.datamodels.vocabulary import IS_A, OBO_PURL, BIOPORTAL_PURL

NC_NAME = str
PREFIX_MAP = Dict[NC_NAME, URI]
RELATIONSHIP_MAP = Dict[PRED_CURIE, List[CURIE]]
ALIAS_MAP = Dict[PRED_CURIE, List[str]]
METADATA_MAP = Dict[PRED_CURIE, List[str]]
RELATIONSHIP = Tuple[CURIE, PRED_CURIE, CURIE]





class BasicOntologyInterface(OntologyInterface, ABC):
    """
    Basic lookup operations on ontologies

    No Object Model is used - input and payloads are simple scalars, dicts, and lists

    This presents an intentionally highly simplistic, lossy view of an ontology. It
    is intended to fit the "80% of uses" scenario - for more advanced uses, use one of the
    more specialized subclasses

    Basic concepts:

    - An ontology is a collection of entities
    - Entities are typically classes (terms), but they may be other types
    - An entity MUST BE uniquely identifier by a CURIE
    - Each entity SHOULD have a single name
    - Entities may have hierarchical parents and children
    - Entities may have simple relationships to other terms

        - A relationship is a simple triple (in OWL terms, these may represent existential relationships)
        - A CURIE is used for the predicate
    - Entities may have aliases (synonyms)

       - these MAY be typed (scoped)

    - Entities have simple key-value metadata associated with them, where keys are open-ended

    Out of scope:

    - Many OWL axioms including:

        - Disjointness
        - Equivalence (including "logical definitions")
        - Property characteristics

    - All OWL expressions

        - but note that simple SomeValuesFrom expressions are included as relationships

    - Information about specific assignments ("axiom annotations" in OWL terms)

        - this includes attribution for synonyms, definitions, ...

    - A fixed datamodel for metadata, beyond core minimal information

    CURIEs are used as the default means of identifying entities.

    In some cases there may not be a defined CURIE form for an entity, such as an ontology,
    in which case URIs may be used

    """


    def get_prefix_map(self) -> PREFIX_MAP:
        """
        Returns all prefixes known to the resource plus their URI expansion

        :return:
        """
        raise NotImplementedError

    def curie_to_uri(self, curie: CURIE, strict=False) -> Optional[URI]:
        """
        Expands a CURIE

        :param curie:
        :param strict:
        :return:
        """
        if curie.startswith('http'):
            return curie
        pm = self.get_prefix_map()
        pfx, local_id = curie.split(':')
        if pfx in pm:
            return f'{pm[pfx]}{local_id}'
        else:
            # TODO: not hardcode
            return f'{OBO_PURL}{pfx}_{local_id}'

    def uri_to_curie(self, uri: URI, strict=True) -> Optional[CURIE]:
        """
        Contracts a URI

        If strict conditions hold, then no URI can map to more than one CURIE (i.e one URI base should not start with another)

        :param uri:
        :param strict:
        :return:
        """
        pm = self.get_prefix_map()
        for k, v in pm.items():
            if uri.startswith(v):
                return uri.replace(v, f'{k}:')
        if uri.startswith(OBO_PURL):
            # TODO: do not hardcode OBO purl behavior
            uri = uri.replace(f'{OBO_PURL}', "")
            return uri.replace('_', ':')
        if uri.startswith(BIOPORTAL_PURL):
            # TODO: do not hardcode OBO purl behavior
            uri = uri.replace(f'{BIOPORTAL_PURL}', "")
            return uri.replace('_', ':')
        return uri


    def all_ontology_curies(self) -> Iterable[CURIE]:
        """
        returns iterator over all known ontology CURIEs

        Many OntologyInterfaces will wrap a single ontology, others will wrap multiple.
        Even when a single ontology is wrapped, there may be multiple ontologies included as imports

        :return: iterator
        """
        raise NotImplementedError

    def all_entity_curies(self) -> Iterable[CURIE]:
        """
        returns iterator over all known entity CURIEs

        :return: iterator
        """
        raise NotImplementedError

    def all_relationships(self) -> Iterable[RELATIONSHIP]:
        """
        returns iterator over all known relationships

        :return:
        """
        for curie in self.all_entity_curies():
            for pred, fillers in self.get_outgoing_relationships_by_curie(curie).items():
                for filler in fillers:
                    yield curie, pred, filler

    def all_subset_curies(self) -> Iterable[SUBSET_CURIE]:
        """
        returns iterator over all known subset CURIEs

        :return: iterator
        """
        raise NotImplementedError

    def curies_by_subset(self, subset: SUBSET_CURIE) -> Iterable[CURIE]:
        """
        returns iterator over all CURIEs belonging to a subset

        :return: iterator
        """
        raise NotImplementedError

    def get_label_by_curie(self, curie: CURIE) -> Optional[str]:
        """
        fetches the unique label for a CURIE

        The CURIE may be for a class, individual, property, or ontology

        :param curie:
        :return:
        """
        raise NotImplementedError

    def get_labels_for_curies(self, curies: Iterable[CURIE]) -> Iterable[Tuple[CURIE, str]]:
        """
        fetches the unique label for a CURIE

        The CURIE may be for a class, individual, property, or ontology

        :param curie:
        :return:
        """
        # default implementation: may be overridden for efficiency
        for curie in curies:
            yield [curie, self.get_label_by_curie(curie)]

    def set_label_for_curie(self, curie: CURIE, label: str) -> bool:
        """
        updates the label

        :param curie:
        :param label:
        :return:
        """
        raise NotImplementedError

    def get_curies_by_label(self, label: str) -> List[CURIE]:
        """
        Fetches all curies with a given label

        This SHOULD return maximum one CURIE but there are circumstances where multiple CURIEs
        may share a label

        :param label:
        :return:
        """
        raise NotImplementedError()

    def get_parents_by_curie(self, curie: CURIE, isa_only: bool = False) -> List[CURIE]:
        """
        Returns all hierarchical parents

        The definition of hierarchical parent depends on the provider. For example, in OLS,
        this will return part-of parents for GO, as well as is-a

        This only returns Named Entities; i.e. if an RDF source is wrapped, this will NOT return blank nodes

        To get all relationships, see :ref:`get_outgoing_relationships_by_curie`

        A fresh map is created each invocation

        :param curie:
        :param isa_only: restrict hierarchical parents to isa only
        :return:
        """
        return self.get_outgoing_relationships_by_curie(curie)[IS_A]

    def get_outgoing_relationships_by_curie(self, curie: CURIE) -> RELATIONSHIP_MAP:
        """
        The return relationship map is keyed by relationship type, where the values
        are the 'parents' or fillers

        OWL formulation:

         - is_a: {P : SubClassOf(C P), class(P)}
         - R: {P : SubClassOf(C ObjectSomeValuesFrom( RP), class(P), property(P)}

        :param curie: the 'child' term
        :return:
        """
        raise NotImplementedError()

    def get_incoming_relationships_by_curie(self, curie: CURIE) -> RELATIONSHIP_MAP:
        """

        :param curie:
        :return:
        """
        raise NotImplementedError()

    def get_definition_by_curie(self, curie: CURIE) -> Optional[str]:
        """

        :param curie:
        :return:x`
        """
        raise NotImplementedError()

    def get_simple_mappings_by_curie(self, curie: CURIE) -> Iterable[Tuple[PRED_CURIE, CURIE]]:
        """

        :param curie:
        :return:
        """
        raise NotImplementedError()

    def aliases_by_curie(self, curie: CURIE) -> List[str]:
        """
        All aliases/synonyms for a given CURIE

        This lumps together all synonym categories (scopes)

        :param curie:
        :return:
        """
        aliases = set()
        for v in self.alias_map_by_curie(curie).values():
            aliases.update(set(v))
        return list(aliases)

    def alias_map_by_curie(self, curie: CURIE) -> ALIAS_MAP:
        """
        Returns aliases keyed by alias type (scope in OBO terms)

        :param curie:
        :return:
        """
        raise NotImplementedError

    def metadata_map_by_curie(self, curie: CURIE) -> METADATA_MAP:
        """
        Returns a dictionary keyed by property predicate, with a list of zero or more values,
        where each key corresponds to any of a set of open-ended metadata predicates

        This is a simplified version of the metadata model in many ontologies

        :param curie:
        :return:
        """
        raise NotImplementedError

    def create_entity(self, curie: CURIE, label: str = None, relationships: RELATIONSHIP_MAP = None) -> CURIE:
        """
        Adds an entity to the resource

        :param curie:
        :param label:
        :param relationships:
        :return:
        """
        raise NotImplementedError




"""
Command Line Interface to OAK
-----------------------------

Executed using "runoak" command
"""
import logging
import os
import subprocess
import sys
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Sequence, TextIO, Tuple, Any, Type

import click
import rdflib
from linkml_runtime.dumpers import yaml_dumper, json_dumper
from oaklib.datamodels.validation_datamodel import ValidationConfiguration
from oaklib.implementations.aggregator.aggregator_implementation import AggregatorImplementation
from oaklib.implementations.sqldb.sql_implementation import SqlImplementation
from oaklib.interfaces import BasicOntologyInterface, OntologyInterface, ValidatorInterface, SubsetterInterface
from oaklib.interfaces.mapping_provider_interface import MappingProviderInterface
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.interfaces.rdf_interface import RdfInterface
from oaklib.interfaces.search_interface import SearchInterface
from oaklib.interfaces.text_annotator_interface import TextAnnotatorInterface
from oaklib.io.streaming_csv_writer import StreamingCsvWriter
from oaklib.io.streaming_yaml_writer import StreamingYamlWriter
from oaklib.resource import OntologyResource
from oaklib.selector import get_resource_from_shorthand, get_implementation_from_shorthand
from oaklib.types import PRED_CURIE
from oaklib.utilities.apikey_manager import set_apikey_value
from oaklib.utilities.iterator_utils import chunk
from oaklib.utilities.lexical.lexical_indexer import create_lexical_index, save_lexical_index, lexical_index_to_sssom, \
    load_lexical_index, load_mapping_rules, add_labels_from_uris
from oaklib.utilities.mapping.sssom_utils import StreamingSssomWriter
from oaklib.utilities.obograph_utils import draw_graph, graph_to_image, default_stylemap_path
import sssom.writers as sssom_writers
from oaklib.datamodels.vocabulary import IS_A, PART_OF, EQUIVALENT_CLASS
from oaklib.utilities.subsets.slimmer_utils import roll_up_to_named_subset
from oaklib.utilities.taxon.taxon_constraint_utils import get_term_with_taxon_constraints, test_candidate_taxon_constraint, parse_gain_loss_file
import oaklib.datamodels.taxon_constraints as tcdm

@dataclass
class Settings:
    impl: Any = None


settings = Settings()

input_option = click.option(
    "-i",
    "--input",
    help="path to input implementation specification."
)
add_option = click.option(
    "-a",
    "--add",
    multiple=True,
    help="additional implementation specification."
)
input_type_option = click.option(
    "-I",
    "--input-type",
    help="Input type."
)
output_option = click.option(
    "-o",
    "--output",
    type=click.File(mode="w"),
    default=sys.stdout,
    help="Output file, e.g. obo file"
)
output_type_option = click.option(
    "-O",
    "--output-type",
    help=f'Desired output type',
)
predicates_option = click.option(
    "-p",
    "--predicates",
    help="A comma-separated list of predicates"
)

def _process_predicates_arg(preds_str: str) -> List[PRED_CURIE]:
    if preds_str is None:
        return None
    inputs = preds_str.split(',')
    preds = [_shorthand_to_pred_curie(p) for p in inputs]
    return preds

def _shorthand_to_pred_curie(shorthand: str) -> PRED_CURIE:
    if shorthand == 'i':
        return IS_A
    elif shorthand == 'p':
        return PART_OF
    elif shorthand == 'e':
        return EQUIVALENT_CLASS
    else:
        return shorthand


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@input_option
@add_option
def main(verbose: int, quiet: bool, input: str, add: List):
    """Run the oaklib Command Line.

    A subcommand must be passed - for example: ancestors, terms, ...

    Most commands require an input ontology to be specified:

        runoak -i <INPUT SPECIFICATION> SUBCOMMAND <SUBCOMMAND OPTIONS AND ARGUMENTS>

    Get help on any command, e.g:

        runoak viz -h
    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    if quiet:
        logging.basicConfig(level=logging.ERROR)
    resource = OntologyResource()
    resource.slug = input

    if input:
        impl_class: Type[OntologyInterface]
        resource = get_resource_from_shorthand(input)
        impl_class = resource.implementation_class
        logging.info(f'RESOURCE={resource}')
        settings.impl = impl_class(resource)
    if add:
        impls = [get_implementation_from_shorthand(d) for d in add]
        if settings.impl:
            impls = [settings.impl] + impls
        settings.impl = AggregatorImplementation(implementations=impls)


@main.command()
@click.argument("terms", nargs=-1)
@output_option
def search(terms, output: str):
    """
    Searches ontology for entities that have a label, alias, or other property matching a search term.

    Example:

        runoak -i uberon.obo search limb

    This uses the Pronto implementation to load uberon from disk, and does a basic substring
    search over the labels and synonyms - results are not ranked

    Bioportal (all ontologies):

        runoak -i bioportal: search limb

    (You need to set your API key first)

    This uses the Bioportal API to search over a broad set of ontologies, returning a ranked list
    ranked by relevance. There may be many results, the results are streamed, do ctrl^C to stop

    Ubergraph (all ontologies):

        runoak -i ubergraph: search limb

    Ubergraph (one ontology):

        runoak -i ubergraph:uberon search limb

    """
    impl = settings.impl
    if isinstance(impl, SearchInterface):
        for t in terms:
            for curie_it in chunk(impl.basic_search(t)):
                logging.info('** Next chunk:')
                for curie, label in impl.get_labels_for_curies(curie_it):
                    print(f'{curie} ! {label}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')


@main.command()
@output_option
def all_subsets(output: str):
    """
    Shows all subsets

    Example:
        runoak -i obolibrary:go.obo all-subsets
    """
    impl = settings.impl
    if isinstance(impl, BasicOntologyInterface):
        for subset in impl.all_subset_curies():
            print(f'{subset} ! {impl.get_label_by_curie(subset)}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.argument("subset")
@output_option
def list_subset(subset, output: str):
    """
    Shows IDs in a given subset

    Example:
        runoak -i obolibrary:go.obo list-subset goslim_generic

    Example:
        oak -i sqlite:notebooks/input/go.db list-subset goslim_agr
    """
    impl = settings.impl
    if isinstance(impl, BasicOntologyInterface):
        for curie in impl.curies_by_subset(subset):
            print(f'{curie} ! {impl.get_label_by_curie(curie)}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.argument("words", nargs=-1)
@output_option
def annotate(words, output: str):
    """
    Annotate a piece of text using a Named Entity Recognition annotation

    Example:
        runoak -i bioportal: annotate "enlarged nucleus in T-cells from peripheral blood"
    """
    impl = settings.impl
    text = ' '.join(words)
    if isinstance(impl, TextAnnotatorInterface):
        for ann in impl.annotate_text(text):
            print(yaml_dumper.dumps(ann))
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')


@main.command()
@click.option("--view/--no-view",
              default=True,
              show_default=True,
              help="if view is set then open the image after rendering")
@click.option("--down/--no-down",
              default=False,
              show_default=True,
              help="traverse down")
@click.option("--gap-fill/--no-gap-fill",
              default=False,
              show_default=True,
              help="If set then find the minimal graph that spans all input curies")
@click.option('-S', '--stylemap',
              help='a json file to configure visualization. See https://berkeleybop.github.io/kgviz-model/')
@click.option('-C', '--configure',
              help='overrides for stylemap, specified as yaml. E.g. `-C "styles: [filled, rounded]" `')
@click.argument("terms", nargs=-1)
@predicates_option
@output_type_option
# TODO: the main output option uses a filelike object
@click.option('-o', '--output',
              help="Path to output file")
#@output_option
def viz(terms, predicates, down, gap_fill, view, stylemap, configure, output_type: str, output: str):
    """
    Visualizing an ancestor graph using obographviz

    Note the implementation must implement :class:`.OboGraphInterface`

    This requires that `obographviz <https://github.com/cmungall/obographviz>`_ is installed.

    Example:

        runoak -i sqlite:cl.db viz CL:4023094

    Same query on ubergraph:

        runoak -i ubergraph: viz CL:4023094

    Example, showing only is-a:

        runoak -i sqlite:cl.db viz CL:4023094 -p i

    Example, showing only is-a and part-of, to include Uberon:

        runoak -i sqlite:cl.db viz CL:4023094 -p i,p

    As above, including develops-from:

        runoak -i sqlite:cl.db viz CL:4023094 -p i,p,RO:0002202
    """
    impl = settings.impl
    if isinstance(impl, OboGraphInterface):
        terms_expanded = []
        for term in terms:
            if term.startswith('in_subset='):
                term = term.replace('in_subset=', '')
                terms_expanded += list(impl.curies_by_subset(term))
            else:
                terms_expanded.append(term)
        if stylemap is None:
            stylemap = default_stylemap_path()
        actual_predicates = _process_predicates_arg(predicates)
        if isinstance(impl, SearchInterface):
            curies = list(impl.multiterm_search(terms_expanded))
        else:
            logging.warning(f'Search not implemented: using direct inputs')
            curies = terms_expanded
        if down:
            graph = impl.subgraph(curies, predicates=actual_predicates)
        elif gap_fill:
            logging.info(f'Using gap-fill strategy')
            if isinstance(impl, SubsetterInterface):
                rels = impl.gap_fill_relationships(curies, predicates=actual_predicates)
                if isinstance(impl, OboGraphInterface):
                    graph = impl.relationships_to_graph(rels)
                else:
                    assert False
            else:
                raise NotImplementedError(f'{impl} needs to implement Subsetter for --gap-fill')
        else:
            graph = impl.ancestor_graph(curies, predicates=actual_predicates)
        logging.info(f'Drawing graph seeded from {curies}')
        if output_type == 'json':
            if output:
                json_dumper.dump(graph, to_file=output, inject_type=False)
            else:
                print(json_dumper.dumps(graph))
        elif output_type == 'yaml':
            if output:
                yaml_dumper.dump(graph, to_file=output, inject_type=False)
            else:
                print(yaml_dumper.dumps(graph))
        else:
            imgfile = graph_to_image(graph, seeds=curies, stylemap=stylemap, configure=configure, imgfile=output)
            if view:
                subprocess.run(['open', imgfile])
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.argument("terms", nargs=-1)
@predicates_option
@output_option
def ancestors(terms, predicates, output: str):
    """
    List all ancestors

    Example:

        runoak -i cl.owl ancestors CL:4023094

    Note that ancestors is by default over ALL relationship types

    Constrained to is-a and part-of:

        runoak -i cl.owl ancestors CL:4023094 -p i,BFO:0000050

    Same, on ubergraph:

        runoak -i ubergraph: ancestors CL:4023094 -p i,BFO:0000050

    Search terms can also be used:

        runoak -i cl.owl ancestors 'goblet cell'

    Multiple terms can be passed:

        runoak -i sqlite:go.db ancestors GO:0005773 GO:0005737 -p i,p
    """
    impl = settings.impl
    if isinstance(impl, OboGraphInterface) and isinstance(impl, SearchInterface):
        actual_predicates = _process_predicates_arg(predicates)
        curies = list(impl.multiterm_search(terms))
        logging.info(f'Ancestor seed: {curies}')
        graph = impl.ancestor_graph(curies, predicates=actual_predicates)
        for n in graph.nodes:
            print(f'{n.id} ! {n.label}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')


@main.command()
@click.argument("terms", nargs=-1)
@predicates_option
@output_option
def descendants(terms, predicates, output: str):
    """
    List all descendants of a term

    See examples for 'ancestors' command
    """
    impl = settings.impl
    if isinstance(impl, OboGraphInterface):
        actual_predicates = _process_predicates_arg(predicates)
        graph = impl.descendant_graph(list(impl.multiterm_search(terms)), predicates=actual_predicates)
        for n in graph.nodes:
            print(f'{n.id} ! {n.label}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.argument("terms", nargs=-1)
@predicates_option
@output_option
@output_type_option
def extract_triples(terms, predicates, output, output_type: str = 'ttl'):
    """
    Extracts a subontology as triples

    Currently the only endpoint to implement this is ubergraph. Ontobee seems
    to have performance issues with the query

    This will be supported in the SqlDatabase/Sqlite endpoint soon

    Example:

        runoak -v -i ubergraph: extract-triples GO:0005635 CL:0000099 -o test.ttl -O ttl

    """
    impl = settings.impl
    if isinstance(impl, RdfInterface):
        actual_predicates = _process_predicates_arg(predicates)
        g = rdflib.Graph()
        for t in impl.extract_triples(terms, map_to_curies=False):
            logging.info(f'Triple: {t}')
            g.add(t)
        output.write(g.serialize())
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.argument("terms", nargs=-1)
@output_option
def info(terms, output: str):
    """
    Show info on terms

    TODO: currenly this only shows the label

    Example:
        runoak -i cl.owl info CL:4023094
    """
    impl = settings.impl
    if isinstance(impl, BasicOntologyInterface):
        curies = list(impl.multiterm_search(terms))
        for curie in curies:
            print(f'{curie} ! {impl.get_label_by_curie(curie)}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
#@output_option
@click.option('-o', '--output',
              help="output file")
@output_type_option
def convert(output: str, output_type):
    """
    TODO
    """
    impl = settings.impl
    if isinstance(impl, BasicOntologyInterface):
        resource = get_resource_from_shorthand(output, format=output_type)
        curies = impl.store(resource)
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.argument("terms", nargs=-1)
@output_option
def relationships(terms, output: str):
    """
    Show relationships for terms

    Example:
        runoak -i cl.owl relationships CL:4023094
    """
    impl = settings.impl
    if isinstance(impl, BasicOntologyInterface):
        curies = list(impl.multiterm_search(terms))
        for curie in curies:
            print(f'{curie} ! {impl.get_label_by_curie(curie)}')
            for pred, fillers in impl.get_outgoing_relationships_by_curie(curie).items():
                print(f'  PRED: {pred} ! {impl.get_label_by_curie(pred)}')
                for filler in fillers:
                    print(f'    * {filler} ! {impl.get_label_by_curie(filler)}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@output_option
def terms(output: str):
    """
    List all terms in the ontology
    """
    impl = settings.impl
    if isinstance(impl, BasicOntologyInterface):
        for curie in impl.all_entity_curies():
            print(f'{curie} ! {impl.get_label_by_curie(curie)}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@output_option
def mappings(output):
    """
    List all SSSOM mappings in the ontology
    """
    impl = settings.impl
    writer = StreamingYamlWriter(output)
    if isinstance(impl, MappingProviderInterface):
        for mapping in impl.all_sssom_mappings():
            writer.emit(mapping)
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@output_option
@output_type_option
@click.argument('curies', nargs=-1)
def term_mappings(curies, output, output_type):
    """
    List all SSSOM mappings for a term or terms

    Example:

        runoak -i bioportal: term-mappings  UBERON:0002101

    Example:

        runoak -i ols: term-mappings  UBERON:0002101

    Example:

        runoak -i db/uberon.db term-mappings  UBERON:0002101
    """
    impl = settings.impl
    if output_type is None or output_type == 'yaml':
        writer = StreamingYamlWriter(output)
    elif output_type == 'csv':
        writer = StreamingCsvWriter(output)
    elif output_type == 'sssom':
        writer = StreamingSssomWriter(output)
    else:
        raise ValueError(f'No such format: {output_type}')
    if isinstance(impl, MappingProviderInterface):
        for curie in curies:
            for mapping in impl.get_sssom_mappings_by_curie(curie):
                writer.emit(mapping)
        writer.close()
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@output_option
def aliases(output):
    """
    List all aliases in the ontology
    """
    impl = settings.impl
    writer = StreamingCsvWriter(output)
    if isinstance(impl, BasicOntologyInterface):
        for curie in impl.all_entity_curies():
            for pred, aliases in impl.alias_map_by_curie(curie).items():
                for alias in aliases:
                    writer.emit(dict(curie=curie, pred=pred, alias=alias))
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@output_option
@click.argument('subsets', nargs=-1)
def subset_rollups(subsets: list, output):
    """
    For each subset provide a mapping of each term in the ontology to a subset
    """
    impl = settings.impl
    #writer = StreamingCsvWriter(output)
    if isinstance(impl, OboGraphInterface):
        impl.enable_transitive_query_cache()
        term_curies = list(impl.all_entity_curies())
        output.write("\t".join(['subset', 'term', 'subset_term']))
        if len(subsets) == 0:
            subsets = list(impl.all_subset_curies())
        for subset in subsets:
            logging.info(f'Subset={subset}')
            m = roll_up_to_named_subset(impl, subset, term_curies, predicates=[IS_A, PART_OF])
            for term, mapped_to in m.items():
                for tgt in mapped_to:
                    output.write("\t".join([subset, term, tgt]))
                    output.write("\n")
                    #writer.emit(dict(subset=subset, term=term, subset_term=tgt))
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')



@main.command()
@output_option
def axioms(output: str):
    """
    List all axioms

    TODO: this is a placeholder -- will be added when we add funowl
    """
    impl = settings.impl
    #if isinstance(impl, OwlInterface):
    #    for axiom in impl.axioms():
    #        print(f'{axiom}')
    #else:
    #    raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')
    raise NotImplementedError

@main.command()
@output_option
@predicates_option
@click.option('-A/--no-A',
              '--all/--no-all',
              default=False,
              show_default=True,
              help="if specified then perform for all terms")
@click.option('--include-redundant/--no-include-redundant',
              default=False,
              show_default=True,
              help="if specified then include redundant taxon constraints from ancestral subjects")
@click.argument('curies', nargs=-1)
def taxon_constraints(curies: list, all: bool, include_redundant: bool, predicates: List, output):
    """
    List all taxon constraints for a term
    """
    impl = settings.impl
    writer = StreamingYamlWriter(output)
    if all:
        if curies:
            raise ValueError(f'Do not specify explicit curies with --all option')
        curies = [curie for curie in impl.all_entity_curies() if impl.get_label_by_curie(curie)]
    if isinstance(impl, OboGraphInterface):
        impl.enable_transitive_query_cache()
        for curie in curies:
            st = get_term_with_taxon_constraints(impl, curie, include_redundant=include_redundant, predicates=predicates)
            writer.emit(st)
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')


@main.command()
@click.option('-E', '--evolution-file',
              help="path to file containing gains and losses")
@output_option
@predicates_option
@click.argument('constraints', nargs=-1)
def add_taxon_constraints(constraints, evolution_file, predicates: List, output):
    """
    Try candidate taxon constraint
    """
    impl = settings.impl
    writer = StreamingYamlWriter(output)
    curr = None
    sts = []
    st = None
    while len(constraints) > 0:
        nxt = constraints[0]
        constraints = constraints[1:]
        if nxt == '.':
            st = None
            curr = None
            continue
        if st is None:
            st = tcdm.SubjectTerm(nxt, label=impl.get_label_by_curie(nxt))
            curr = st.only_in
            sts.append(st)
        else:
            if nxt.lower().startswith('only'):
                curr = st.only_in
            elif nxt.lower().startswith('never'):
                curr = st.never_in
            elif nxt.lower().startswith('present'):
                curr = st.present_in
            else:
                tc = tcdm.TaxonConstraint(taxon=tcdm.Taxon(nxt))
                curr.append(tc)
    if evolution_file is not None:
        with open(evolution_file) as file:
            sts += list(parse_gain_loss_file(file))
    if isinstance(impl, OboGraphInterface):
        impl.enable_transitive_query_cache()
        for st in sts:
            try:
                st = test_candidate_taxon_constraint(impl, st, predicates=predicates)
                writer.emit(st)
            except ValueError as e:
                logging.error(f'Error with TC: {e}')
                st.description = 'PROBLEM'
                writer.emit(st)
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.option('--cutoff',
              default=50,
              show_default=True,
              help="maximum results to report for any (type, predicate) pair")
@output_option
def validate(output: str, cutoff: int):
    """
    Validate an ontology against ontology metadata
    """
    writer = StreamingCsvWriter(output)
    impl = settings.impl
    if isinstance(impl, ValidatorInterface):
        counts = defaultdict(int)
        for result in impl.validate():
            key = (result.type, result.predicate)
            n = counts[key]
            n += 1
            counts[key] = n
            if n % 1000 == 0:
                logging.info(f'Reached {n} results with {key}')
            if n == cutoff:
                print(f'**TRUNCATING RESULTS FOR {key} at {cutoff}')
            elif n < cutoff:
                writer.emit(result)
                print(yaml_dumper.dumps(result))
        for k, v in counts.items():
            print(f'{k}:: {v}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')


@main.command()
@click.option('--cutoff',
              default=50,
              show_default=True,
              help="maximum results to report for any (type, predicate) pair")
@click.option('-s', '--schema',
              help="Path to schema (if you want to override the bundled OMO schema)")
@click.argument("dbs", nargs=-1)
@output_option
def validate_multiple(dbs, output, schema, cutoff: int):
    """
    Validate an ontology against ontology metadata
    """
    writer = StreamingCsvWriter(output)
    config = ValidationConfiguration()
    if schema:
        config.schema_path = schema
    for db in dbs:
        try:
            path = Path(db).absolute()
            print(f'PATH={path}')
            resource = OntologyResource(slug=f'sqlite:///{str(path)}')
            impl = SqlImplementation(resource)
            counts = defaultdict(int)
            for result in impl.validate(configuration=config):
                result.source = f'sqlite:{db}'
                key = (result.type, result.predicate)
                n = counts[key]
                n += 1
                counts[key] = n
                if n % 1000 == 0:
                    logging.info(f'Reached {n} results with {key}')
                if n == cutoff:
                    print(f'**TRUNCATING RESULTS FOR {key} at {cutoff}')
                elif n < cutoff:
                    try:
                        print(yaml_dumper.dumps(result))
                        writer.emit(result)
                    except ValueError as e:
                        logging.error(e)
                        logging.error(f'Could not dump {result} -- bad identifier?')
        except Exception as e:
            logging.error(e)
            logging.error(f'Problem with db')
        for k, v in counts.items():
            print(f'{k}:: {v}')


@main.command()
@output_option
def check_definitions(output: str):
    """
    Check definitions
    """
    impl = settings.impl
    if isinstance(impl, ValidatorInterface):
        for curie in impl.term_curies_without_definitions():
            print(f'NO DEFINITION: {curie} ! {impl.get_label_by_curie(curie)}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')



@main.command()
@click.argument("curies", nargs=-1)
@output_option
def subset(curies, output: str):
    """
    Extracts a subset

    TODO: INCOMPLETE
    """
    impl = settings.impl
    if isinstance(impl, SubsetterInterface):
        subont = impl.extract_subset_ontology(curies)
        print(f'TODO: {subont}')
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')

@main.command()
@click.option('--endpoint',
              '-e')
@click.argument("keyval")
def set_apikey(endpoint, keyval):
    """
    Sets an API key

    Example:
        oak set-apikey -e bioportal MY-KEY-VALUE
    """
    set_apikey_value(endpoint, keyval)


@main.command()
@click.option("--lexical-index-file", "-L",
              help="path to lexical index. This is recreated each time unless --no-recreate is passed")
@click.option("--rules-file", "-R",
              help="path to rules file. Conforms to rules_datamodel.")
@click.option("--add-labels/--no-add-labels",
              default=False,
              show_default=True,
              help="Populate empty labels with URI fragments or CURIE local IDs, for ontologies that use semantic IDs")
@click.option("--recreate/--no-recreate",
              default=True,
              show_default=True,
              help="if true and lexical index is specified, always recreate, otherwise load from index")
@output_option
def lexmatch(output, recreate, rules_file, lexical_index_file, add_labels):
    """
    Generates lexical index and mappings

    See :ref:`.lexical_index_to_sssom`

    Examples:
        lexmatch -i foo.obo -o foo.sssom.tsv

    Outputting intermediate index:
        lexmatch -i foo.obo -L foo.index.yaml -o foo.sssom.tsv

    Using custom rules:
        lexmatch -i foo.obo -R match_rules.yaml -L foo.index.yaml -o foo.sssom.tsv
    """
    impl = settings.impl
    if rules_file:
        ruleset = load_mapping_rules(rules_file)
    else:
        ruleset = None
    if isinstance(impl, BasicOntologyInterface):
        if add_labels:
            add_labels_from_uris(impl)
        if not recreate and Path(lexical_index_file).exists():
            ix = load_lexical_index(lexical_index_file)
        else:
            ix = create_lexical_index(impl)
        if lexical_index_file:
            if recreate:
                save_lexical_index(ix, lexical_index_file)
        msdf = lexical_index_to_sssom(impl, ix, ruleset=ruleset)
        sssom_writers.write_table(msdf, output)
    else:
        raise NotImplementedError(f'Cannot execute this using {impl} of type {type(impl)}')


if __name__ == "__main__":
    main()



from abc import ABC
from typing import Dict, List, Tuple, Iterable

from oaklib.interfaces.basic_ontology_interface import BasicOntologyInterface
import sssom
from oaklib.types import CURIE


class MappingProviderInterface(BasicOntologyInterface, ABC):
    """
    An ontology provider that provides SSSOM mappings

    TODO: move code from mapping-walker
    """

    def all_sssom_mappings(self) -> Iterable[sssom.Mapping]:
        """
        All SSSOM mappings in the ontology

        :return:
        """
        for curie in self.all_entity_curies():
            for m in self.get_sssom_mappings_by_curie(curie):
                yield m

    def get_sssom_mappings_by_curie(self, curie: CURIE) -> Iterable[sssom.Mapping]:
        """
        All SSSOM mappings about a curie

        :param curie:
        :return:
        """
        raise NotImplementedError

    def get_transitive_mappings_by_curie(self, curie: CURIE) -> Iterable[sssom.Mapping]:
        raise NotImplementedError

